/* ////////////////////////////////////////////////////////////////////////////
//
//                  INTEL CORPORATION PROPRIETARY INFORMATION
//     This software is supplied under the terms of a license agreement or
//     nondisclosure agreement with Intel Corporation and may not be copied
//     or disclosed except in accordance with the terms of that agreement.
//          Copyright(c) 2009 Intel Corporation. All Rights Reserved.
*/

#include <stdlib.h>
#include <stdio.h>

#include "ippdefs.h"
#include "ippcore.h"
#include "ipps.h"
#include "ippi.h"
#include "ippcv.h"


static void ipp_version(void)
{
  const IppLibraryVersion* version;

  printf("Intel(R) Integrated Performance Primitives\n");

  /* ippCore version info */
  version = ippGetLibVersion();

  printf("  %s, %s {%d.%d.%d.%d}, build date %s\n",
    version->Name,
    version->Version,
    version->major,version->minor,version->majorBuild,version->build,
    version->BuildDate);

  /* ippSP version info */
  version = ippsGetLibVersion();

  printf("  %s, %s {%d.%d.%d.%d}, build date %s\n",
    version->Name,
    version->Version,
    version->major,version->minor,version->majorBuild,version->build,
    version->BuildDate);

  /* ippIP version info */
  version = ippiGetLibVersion();

  printf("  %s, %s {%d.%d.%d.%d}, build date %s\n",
    version->Name,
    version->Version,
    version->major,version->minor,version->majorBuild,version->build,
    version->BuildDate);

  /* ippCV version info */
  version = ippcvGetLibVersion();

  printf("  %s, %s {%d.%d.%d.%d}, build date %s\n",
    version->Name,
    version->Version,
    version->major,version->minor,version->majorBuild,version->build,
    version->BuildDate);

  return;
} /* ipp_version() */


void pyroptflow(
  const Ipp8u*         prevFrame, // previous frame
        int            prevStep,  // its row step
  const Ipp8u*         nextFrame, // next frame
        int            nextStep,  // its row step
        IppiSize       roiSize,   // frame size
        int            numLevel,  // pyramid level number (5)
        float          rate,      // pyramid rate (2.0f)
        Ipp16s*        pKernel,   // pyramid kernel
        int            kerSize,   // pyramid kernel size (5)
        IppiPoint_32f* prevPt,    // coordinates on previous frame
        IppiPoint_32f* nextPt,    // hint to coordinates on next frame
        Ipp8s*         pStatus,   // result indicators
        Ipp32f*        pError,    // differences
        int            numFeat,   // point number
        int            winSize,   // search window size (41)
        int            numIter,   // iteration number (5)
        float          threshold) // threshold (0.0001f)
{
  IppiPyramid*      pPyr1;
  IppiPyramid*      pPyr2;
  IppiOptFlowPyrLK* pOF;
  IppStatus         status;

  status = ippiPyramidInitAlloc(&pPyr1, numLevel, roiSize, rate);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidInitAlloc: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  status = ippiPyramidInitAlloc(&pPyr2, numLevel, roiSize, rate);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidInitAlloc: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  IppiPyramidDownState_8u_C1R** pState1 = (IppiPyramidDownState_8u_C1R**)&(pPyr1->pState);
  IppiPyramidDownState_8u_C1R** pState2 = (IppiPyramidDownState_8u_C1R**)&(pPyr2->pState);

  Ipp8u**   pImg1 = pPyr1->pImage;
  Ipp8u**   pImg2 = pPyr2->pImage;

  int*      pStep1 = pPyr1->pStep;
  int*      pStep2 = pPyr2->pStep;

  IppiSize* pRoi1 = pPyr1->pRoi;
  IppiSize* pRoi2 = pPyr2->pRoi;

  IppHintAlgorithm hint = ippAlgHintFast;

  int i;
  int level = pPyr1->level;

  status = ippiPyramidLayerDownInitAlloc_8u_C1R(pState1,roiSize,rate,pKernel,kerSize, IPPI_INTER_LINEAR);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidLayerDownInitAlloc_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  status = ippiPyramidLayerDownInitAlloc_8u_C1R(pState2,roiSize,rate,pKernel,kerSize, IPPI_INTER_LINEAR);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidLayerDownInitAlloc_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  pImg1[0] = (Ipp8u*)prevFrame;
  pImg2[0] = (Ipp8u*)nextFrame;

  pStep1[0] = prevStep;
  pStep2[0] = nextStep;

  pRoi1[0] = roiSize;
  pRoi2[0] = roiSize;

  for(i = 1; i <= level; i++)
  {
    pPyr1->pImage[i] = ippiMalloc_8u_C1(pRoi1[i].width,pRoi1[i].height,pStep1+i);
    if(0 == pPyr1->pImage[i])
    {
      printf("ippiMalloc_8u_C1: mem allocation error\n");
      return;
    }

    pPyr2->pImage[i] = ippiMalloc_8u_C1(pRoi2[i].width,pRoi2[i].height,pStep2+i);
    if(0 == pPyr2->pImage[i])
    {
      printf("ippiMalloc_8u_C1: mem allocation error\n");
      return;
    }

    status = ippiPyramidLayerDown_8u_C1R(pImg1[i-1],pStep1[i-1],pRoi1[i-1],pImg1[i],pStep1[i],pRoi1[i],*pState1);
    if(ippStsNoErr != status)
    {
      printf("ippiPyramidLayerDown_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
      return;
    }

    status = ippiPyramidLayerDown_8u_C1R(pImg2[i-1],pStep2[i-1],pRoi2[i-1],pImg2[i],pStep2[i],pRoi2[i],*pState2);
    if(ippStsNoErr != status)
    {
      printf("ippiPyramidLayerDown_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
      return;
    }
  }

  status = ippiOpticalFlowPyrLKInitAlloc_8u_C1R(&pOF,roiSize,winSize,hint);
  if(ippStsNoErr != status)
  {
    printf("ippiOpticalFlowPyrLKInitAlloc_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  status = ippiOpticalFlowPyrLK_8u_C1R(pPyr1,pPyr2,prevPt,nextPt,pStatus,pError,
    numFeat,winSize,numLevel,numIter,threshold,pOF);
  if(ippStsNoErr != status)
  {
    printf("ippiOpticalFlowPyrLK_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  printf("Optical Flow result\n");
  printf("  Features: Prev Point[0] {%1.1f,%1.1f}, Next Point[0] {%1.1f,%1.1f}\n",prevPt[0].x,prevPt[0].y,nextPt[0].x,nextPt[0].y);
//  printf("  Features: Prev Point[1] {%1.1f,%1.1f}, Next Point[1] {%1.1f,%1.1f}\n",prevPt[1].x,prevPt[1].y,nextPt[1].x,nextPt[1].y);
//  printf("  Features: Prev Point[2] {%1.1f,%1.1f}, Next Point[2] {%1.1f,%1.1f}\n",prevPt[2].x,prevPt[2].y,nextPt[2].x,nextPt[2].y);
  printf("  pStatus[0]: %d, pError[0]: %f\n",pStatus[0],pError[0]);
//  printf("  pStatus[1]: %d, pError[1]: %1.2f\n",pStatus[1],pError[1]);
//  printf("  pStatus[2]: %d, pError[2]: %1.2f\n",pStatus[2],pError[2]);

  status = ippiOpticalFlowPyrLKFree_8u_C1R(pOF);
  if(ippStsNoErr != status)
  {
    printf("ippiOpticalFlowPyrLKFree_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  for(i = level; i > 0; i--)
  {
    if(pImg2[i])
      ippiFree(pImg2[i]);

    if(pImg1[i])
      ippiFree(pImg1[i]);
  }

  status = ippiPyramidLayerDownFree_8u_C1R(*pState1);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidLayerDownFree_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  status = ippiPyramidLayerDownFree_8u_C1R(*pState2);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidLayerDownFree_8u_C1R: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  status = ippiPyramidFree(pPyr2);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidFree: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  status = ippiPyramidFree(pPyr1);
  if(ippStsNoErr != status)
  {
    printf("ippiPyramidFree: %d [%s]\n",status,ippGetStatusString(status));
    return;
  }

  return;
} // pyroptflow()


#define WIDTH        8
#define HEIGHT       8
#define NUM_LVL      1
#define NUM_FEA      1
#define WINSIZE      5
#define NUMITER      10
#define TRESHOLD     0.0001f
#define KERNEL_SIZE  5
#define RATE         2.0f


Ipp8u prev[WIDTH*HEIGHT] =
{
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,255,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0
};

Ipp8u next[WIDTH*HEIGHT] =
{
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,255,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0
};

IppiPoint_32f prevPt[NUM_FEA] =
{
//  { 0.0f, 0.0f},
  { 2.0f, 2.0f},
//  { 0.0f, 0.0f}
};

IppiPoint_32f nextPt[NUM_FEA] =
{
//  { 0.0f, 0.0f}, // point not change location
  { 2.0f, 2.0f}, // We don't assume new location
//  { 0.0f, 0.0f}  // point moved up
};

int main(int argc, char** argv)
{
  int    prevStep = WIDTH;
  int    nextStep = HEIGHT;
  Ipp8s  pStatus[NUM_FEA] = { 0 };
  Ipp32f pError[NUM_FEA]  = { 0.0 };
  Ipp16s pKernel[KERNEL_SIZE];
  Ipp8u* prevFrame = &prev[0];
  Ipp8u* nextFrame = &next[0];
  IppiSize roiSize = { WIDTH, HEIGHT };

  // just some kernel
  int scale = 1<<15;
  pKernel[0] = (Ipp16s)(0.5f*(0.5f - 0.375f)*scale);
  pKernel[1] = (Ipp16s)(0.25 * scale);
  pKernel[2] = (Ipp16s)(0.375f*scale);
  pKernel[3] = (Ipp16s)(0.25 * scale);
  pKernel[4] = (Ipp16s)(0.5f*(0.5f - 0.375f)*scale);

  ippStaticInit();

  ipp_version();

  pyroptflow(
    prevFrame,     // previous frame
    prevStep,      // its row step
    nextFrame,     // next frame
    nextStep,      // its row step
    roiSize,       // frame size
    NUM_LVL,       // pyramid level number
    RATE,          // pyramid rate
    &pKernel[0],   // pyramid kernel
    KERNEL_SIZE,   // pyramid kernel size
    &prevPt[0],    // coordinates on previous frame
    &nextPt[0],    // hint to coordinates on next frame
    &pStatus[0],   // result indicators
    &pError[0],    // differences
    NUM_FEA,       // point number
    WINSIZE,       // search window size
    NUMITER,       // iteration number
    TRESHOLD);     // threshold

  return 0;
} // main()

